import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List

from tox.plugin import impl

from ._parse_dependencies import parse_pyproject_toml, parse_setup_cfg, parse_single_requirement

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.session.state import State
    from tox.tox_env.api import EnvConfigSet, ToxEnv


def _write_constrains_file(tox_env: "ToxEnv", dependencies: Dict[str, str], extra_lines: List[str]) -> Path:
    if tox_env.options.min_req_constraints_path:
        base_path = Path(tox_env.options.min_req_constraints_path)
        constrain_file = base_path / "constraints.txt" if base_path.is_dir() else base_path
    elif os.environ.get("TOX_MIN_REQ_CONSTRAINTS", ""):
        base_path = Path(os.environ["TOX_MIN_REQ_CONSTRAINTS"])
        constrain_file = base_path / "constraints.txt" if base_path.is_dir() else base_path
    else:
        constrain_file = tox_env.env_tmp_dir / "constraints.txt"

    with constrain_file.open("w") as f:
        f.write("\n".join(f"{n}=={v}" for n, v in dependencies.items()))
        f.write("\n")
        f.write("\n".join(extra_lines))

    if "PIP_CONSTRAINT" in tox_env.environment_variables:
        tox_env.environment_variables["PIP_CONSTRAINT"] += f" {str(constrain_file)}"
    else:
        tox_env.environment_variables["PIP_CONSTRAINT"] = str(constrain_file)

    return constrain_file


@impl
def tox_on_install(tox_env: "ToxEnv", arguments: Any, section: str, of_type: str) -> None:  # noqa: ARG001, ANN401
    if of_type != "deps" and section != "PythonRun":
        return
    if os.environ.get("MIN_REQ", "0") != "1" and not tox_env.conf["min_req"]:
        return

    project_path = tox_env.core._root  # noqa: SLF001
    python_version = ".".join(str(x) for x in tox_env.base_python.version_info[:2])
    python_full_version = ".".join(str(x) for x in tox_env.base_python.version_info[:3])
    if (project_path / "setup.cfg").exists():
        dependencies = parse_setup_cfg(project_path / "setup.cfg", python_version, python_full_version)
    elif (project_path / "pyproject.toml").exists():
        dependencies = parse_pyproject_toml(project_path / "pyproject.toml", python_version, python_full_version)
    else:  # pragma: no cover
        return

    extra_lines = []

    if tox_env.conf["min_req_constraints"]:
        for line in tox_env.conf["min_req_constraints"].split("\n"):
            strip_line = line.strip()
            if strip_line.startswith("-r"):
                if "{project_dir}" in strip_line:
                    strip_line = strip_line.replace("{project_dir}", str(project_path))
                extra_lines.append(strip_line)
            else:
                dependencies.update(parse_single_requirement(strip_line, python_version, python_full_version))

    _write_constrains_file(tox_env, dependencies, extra_lines)


@impl
def tox_add_env_config(env_conf: "EnvConfigSet", state: "State") -> None:  # noqa: ARG001
    env_conf.add_config(
        keys=["min_req"],
        of_type=bool,
        default=False,
        desc="Set to true to use the minimum required version of the dependencies",
    )
    env_conf.add_config(
        keys=["min_req_constraints"],
        of_type=str,
        default="",
        desc="List of additional constraints to use when min_req is set to true, "
        "could override the minimum required version of the dependencies",
    )


@impl
def tox_add_option(parser: "ToxParser") -> None:
    parser.add_argument(
        "--min-req-constraints-path",
        type=str,
        default="",
        help="Path to directory where the constraints.txt file will be created. "
        "If not set, the constraints file will be created in the tox temporary directory. "
        "Because of pip using space as separator, the path should not contain spaces.",
    )



from typing import TYPE_CHECKING

from tox_min_req._parse_dependencies import parse_pyproject_toml, parse_setup_cfg, parse_single_requirement

if TYPE_CHECKING:
    from pathlib import Path

    import pytest


def test_setup_cfg_parse(data_dir: "Path", monkeypatch: "pytest.MonkeyPatch"):
    monkeypatch.setattr("sys.platform", "linux")
    setup_file = data_dir / "setup.cfg"

    constrains = {
        "pytest": "7.0.0",
        "pytest-cov": "2.5",
        "sphinx": "3.0.0",
        "scipy": "1.2.0",
    }

    assert parse_setup_cfg(setup_file, python_version="3.7", python_full_version="3.7.1") == {
        "numpy": "1.16.0",
        **constrains,
    }
    assert parse_setup_cfg(setup_file, python_version="3.8", python_full_version="3.8.3") == {
        "numpy": "1.18.0",
        **constrains,
    }
    monkeypatch.setattr("sys.platform", "win32")
    assert parse_setup_cfg(setup_file, python_version="3.8", python_full_version="3.8.3") == {
        "numpy": "1.18.0",
        "pandas": "0.25.0",
        **constrains,
    }


def test_pyproject_toml_parse(data_dir: "Path", monkeypatch: "pytest.MonkeyPatch"):
    monkeypatch.setattr("sys.platform", "linux")
    pyproject_file = data_dir / "pyproject.toml"

    constrains = {
        "pytest": "7.0.0",
        "pytest-cov": "2.5",
        "sphinx": "3.0.0",
        "scipy": "1.2.0",
    }

    assert parse_pyproject_toml(pyproject_file, python_version="3.7", python_full_version="3.7.1") == {
        "numpy": "1.16.0",
        **constrains,
    }
    assert parse_pyproject_toml(pyproject_file, python_version="3.8", python_full_version="3.8.3") == {
        "numpy": "1.18.0",
        **constrains,
    }
    monkeypatch.setattr("sys.platform", "win32")
    assert parse_pyproject_toml(pyproject_file, python_version="3.8", python_full_version="3.8.3") == {
        "numpy": "1.18.0",
        "pandas": "0.25.0",
        **constrains,
    }


def test_parse_single_requirement():
    p_ver, py_full_ver = "3.10", "3.10.1"
    assert parse_single_requirement("numpy==1.16.0", p_ver, py_full_ver) == {"numpy": "1.16.0"}
    assert parse_single_requirement("numpy>=1.16.0", p_ver, py_full_ver) == {"numpy": "1.16.0"}
    assert parse_single_requirement("numpy>=1.16.0 ; python_version < '3.8'", p_ver, py_full_ver) == {}
    assert parse_single_requirement("numpy[test]>=1.16.0", p_ver, py_full_ver) == {"numpy": "1.16.0"}



import os
import shutil
import sys
from typing import TYPE_CHECKING

import pytest
from tox.pytest import ToxProjectCreator, init_fixture  # noqa: F401

if TYPE_CHECKING:
    from pathlib import Path

tox_ini_template = """
[tox]
envlist = py{env}

[testenv]
extras = test
recreate = True
commands = pytest test_file.py
{extras}
"""

setup_cfg_template = """
[metadata]
name = test_package
version = 0.0.1
author = Grzegorz Bokota

[options]
packages = sample_package
install_requires =
    six>=1.13.0
    click>=7.1.2

[options.extras_require]
test =
    pytest>=7.1.0
    coverage
"""

setup_py_template = """
from setuptools import setup

setup()
"""


pyproject_toml_template_base = """
[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"

[project]
name = "test_package"
version = "0.0.1"
description = "Test package"
dependencies = [
    {deps}
]

[project.optional-dependencies]
test = [
    "pytest>=7.1.0",
]
"""

pyproject_toml_template = pyproject_toml_template_base.format(deps='    "six>=1.13.0",\n    "click>=7.1.2",\n')


test_file_template = """
import pytest
import six
import click

import sample_package

def test_pytest_version():
    assert pytest.__version__ {cmp} "7.1.0"

def test_six_version():
    assert six.__version__ {cmp} "1.13.0"

def test_click_version():
    assert click.__version__ {cmp} "7.1.2"

def test_sample_package():
    assert sample_package.sample_function() == 42
"""


@pytest.fixture(autouse=True)
def _clean_pip_constrains_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("PIP_CONSTRAINT", raising=False)


@pytest.mark.parametrize(("cmp", "req"), [("==", "1"), ("!=", "0")])
def test_setup_cfg_parse(
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
    cmp: str,
    req: str,
) -> None:
    monkeypatch.setenv("MIN_REQ", req)
    env = f"{sys.version_info[0]}{sys.version_info[1]}"
    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=""),
            "setup.cfg": setup_cfg_template,
            "test_file.py": test_file_template.format(cmp=cmp),
            "setup.py": setup_py_template,
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")

    result.assert_success()


@pytest.mark.parametrize(("cmp", "req"), [("==", "1"), ("!=", "0")])
def test_pyproject_toml_parse(
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
    cmp: str,
    req: str,
) -> None:
    monkeypatch.setenv("MIN_REQ", req)
    env = f"{sys.version_info[0]}{sys.version_info[1]}"
    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=""),
            "pyproject.toml": pyproject_toml_template,
            "test_file.py": test_file_template.format(cmp=cmp),
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")

    result.assert_success()


@pytest.mark.parametrize(("cmp", "req"), [("==", "1"), ("!=", "0")])
def test_min_req_config(
    tox_project: ToxProjectCreator,
    data_dir: "Path",
    cmp: str,
    req: str,
) -> None:
    env = f"{sys.version_info[0]}{sys.version_info[1]}"
    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=f"min_req={req}"),
            "pyproject.toml": pyproject_toml_template,
            "test_file.py": test_file_template.format(cmp=cmp),
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")

    result.assert_success()


def test_preserve_constrains(
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
    tmp_path: "Path",
) -> None:
    monkeypatch.setenv("MIN_REQ", "1")
    (tmp_path / "constraints.txt").write_text("coverage==6.5.0")
    monkeypatch.setenv("PIP_CONSTRAINT", str(tmp_path / "constraints.txt"))
    env = f"{sys.version_info[0]}{sys.version_info[1]}"

    test_file_template_ = test_file_template.format(cmp="==")
    test_file_template_ += 'import coverage\ndef test_click_version():\n    assert coverage.__version__ == "6.5.0"\n'
    test_file_template_ += (
        'import os\ndef test_environ():\n    assert len(os.environ["PIP_CONSTRAINT"].split(" ")) == 2\n'
    )
    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=""),
            "setup.cfg": setup_cfg_template,
            "test_file.py": test_file_template_,
            "setup.py": setup_py_template,
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")

    result.assert_success()


def test_additional_constrains(
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
) -> None:
    monkeypatch.setenv("MIN_REQ", "1")
    env = f"{sys.version_info[0]}{sys.version_info[1]}"

    test_file_template_ = test_file_template.format(cmp="==")
    test_file_template_ += 'import coverage\ndef test_click_version():\n    assert coverage.__version__ == "6.5.0"\n'
    test_file_template_ = test_file_template_.replace("1.13.0", "1.14.0")

    req_str = "\n".join(
        f"    {val}"
        for val in ["coverage==6.5.0", "babel==2.6.0", "six==1.14.0", "-r '{project_dir}/constraints_dummy.txt'"]
    )

    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=f"min_req_constraints=\n{req_str}"),
            "setup.cfg": setup_cfg_template,
            "test_file.py": test_file_template_,
            "setup.py": setup_py_template,
            "constraints_dummy.txt": "wheel==0.37.0",
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")

    result.assert_success()


def test_additional_constrains_full_path(
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
    tmp_path: "Path",
) -> None:
    monkeypatch.setenv("MIN_REQ", "1")
    env = f"{sys.version_info[0]}{sys.version_info[1]}"

    test_file_template_ = test_file_template.format(cmp="==")
    test_file_template_ += 'import coverage\ndef test_click_version():\n    assert coverage.__version__ == "6.5.0"\n'
    test_file_template_ = test_file_template_.replace("1.13.0", "1.14.0")

    req_str = "\n".join(
        f"    {val}"
        for val in ["coverage==6.5.0", "babel==2.6.0", "six==1.14.0", f"-r '{tmp_path}/constraints_dummy.txt'"]
    )

    with (tmp_path / "constraints_dummy.txt").open("w") as f:
        f.write("wheel==0.37.0")

    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=f"min_req_constraints=\n{req_str}"),
            "setup.cfg": setup_cfg_template,
            "test_file.py": test_file_template_,
            "setup.py": setup_py_template,
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")

    result.assert_success()


@pytest.mark.parametrize("full_path", [True, False])
@pytest.mark.parametrize("use_cli", [True, False])
def test_constrains_path(
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
    tmp_path: "Path",
    *,
    full_path: bool,
    use_cli: bool,
) -> None:
    env = f"{sys.version_info[0]}{sys.version_info[1]}"
    monkeypatch.setenv("MIN_REQ", "1")
    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=""),
            "pyproject.toml": pyproject_toml_template,
            "test_file.py": test_file_template.format(cmp="=="),
        },
        base=data_dir / "package_data",
    )

    env_path = tmp_path / "env"
    env_path.mkdir()

    if full_path:
        monkeypatch.setenv("TOX_MIN_REQ_CONSTRAINTS", str(env_path / "constraints.txt"))
    else:
        monkeypatch.setenv("TOX_MIN_REQ_CONSTRAINTS", str(env_path))

    expected_file = env_path / "constraints.txt"
    extra_args = []
    if use_cli:
        cli_path = tmp_path / "cli"
        cli_path.mkdir()
        expected_file = cli_path / "constraints.txt"
        extra_args = ["--min-req-constraints-path", str(cli_path / "constraints.txt") if full_path else str(cli_path)]

    result = project.run("run", *extra_args)

    result.assert_success()

    assert expected_file.exists()


test_file_template_six = """
import pytest
import six

import sample_package

def test_pytest_version():
    assert pytest.__version__ == "7.1.0"

def test_six_version():
    assert six.__version__ == "{six_version}"

def test_sample_package():
    assert sample_package.sample_function() == 42
"""


PYTHON_VER_LI = [
    ("3.7", "37", "1.10.0"),
    ("3.8", "38", "1.11.0"),
    ("3.9", "39", "1.12.0"),
    ("3.10", "310", "1.13.0"),
    ("3.11", "311", "1.14.0"),
]


@pytest.mark.parametrize(("python", "env", "target_six_version"), PYTHON_VER_LI)
def test_proper_version_handle(  # noqa: PLR0913
    tox_project: ToxProjectCreator,
    monkeypatch: pytest.MonkeyPatch,
    data_dir: "Path",
    python: str,
    env: str,
    target_six_version: str,
) -> None:
    if not shutil.which(f"python{python}") and os.environ.get("REQUIRE_ALL_TEST", "0") == "0":
        pytest.skip(f"Python {python} is not installed")
    monkeypatch.setenv("MIN_REQ", "1")
    constrains_list = "\n".join(
        f"    'six>={six_version}; python_version == \"{python_version}\"',"
        for python_version, _, six_version in PYTHON_VER_LI
    )
    project = tox_project(
        {
            "tox.ini": tox_ini_template.format(env=env, extras=""),
            "pyproject.toml": pyproject_toml_template_base.format(deps=constrains_list),
            "test_file.py": test_file_template_six.format(six_version=target_six_version),
        },
        base=data_dir / "package_data",
    )

    result = project.run("run")
    result.assert_success()



"""Module to parse the dependencies from the setup.cfg and pyproject.toml files."""

import re
import sys
from configparser import ConfigParser
from pathlib import Path
from typing import Dict, Union

from packaging.requirements import Requirement

if sys.version_info < (3, 11):
    from tomli import loads as toml_loads
else:
    from tomllib import loads as toml_loads

version_constrains = re.compile(r"([a-zA-Z0-9_\-]+)([><=!]+)([0-9\.]+)")

__all__ = ("parse_setup_cfg", "parse_pyproject_toml", "parse_single_requirement")


def parse_single_requirement(line: str, python_version: str, python_full_version: str) -> Dict[str, str]:
    """
    Parse single requirement line. It resolve requirement against current system and provided python version.

    :param line: line with requirement
    :param python_version: major.minor version of python
    :param python_full_version: major.minor.patch version of python
    :return: empty dict if the requirement is not valid or the requirement name and version
    """
    req = Requirement(line)
    if req.marker is not None and not req.marker.evaluate(
        {"python_version": python_version, "python_full_version": python_full_version},
    ):
        return {}
    version_li = [str(x).replace(">=", "").replace("==", "") for x in req.specifier if ">=" in str(x) or "==" in str(x)]
    if version_li:
        return {req.name: version_li[0]}
    return {}


def _parse_setup_cfg_section(section: str, python_version: str, python_full_version: str) -> Dict[str, str]:
    res: Dict[str, str] = {}
    for raw_line in section.splitlines():
        line = raw_line.strip()
        if line.startswith("#") or not line or ">=" not in line:
            continue
        res.update(parse_single_requirement(line, python_version, python_full_version))
    return res


def parse_setup_cfg(path: Union[str, Path], python_version: str, python_full_version: str) -> Dict[str, str]:
    """
    Parse the setup.cfg file and return a dict of the dependencies and their lower version constraints.

    :param path: path to setup.cfg file
    :param python_version: major.minor version of python
    :param python_full_version: major.minor.patch version of python
    :return: dict of the dependencies that fit to environment and their lower version constraints
    """
    config = ConfigParser()
    config.read(path)
    base_constrains = _parse_setup_cfg_section(
        config["options"]["install_requires"],
        python_version,
        python_full_version,
    )
    for extra in config["options.extras_require"]:
        base_constrains.update(
            _parse_setup_cfg_section(config["options.extras_require"][extra], python_version, python_full_version),
        )

    return base_constrains


def parse_pyproject_toml(path: Union[str, Path], python_version: str, python_full_version: str) -> Dict[str, str]:
    """
    Parse the pyproject.toml file and return a dict of the dependencies and their lower version constraints.

    :param path: path to pyproject.toml file
    :param python_version: major.minor version of python
    :param python_full_version: major.minor.patch version of python
    :return: dict of the dependencies that fit to environment and their lower version constraints
    """
    with Path(path).open() as f:
        data = toml_loads(f.read())
    base_constrains: Dict[str, str] = {}
    for line in data["project"]["dependencies"]:
        base_constrains.update(parse_single_requirement(line, python_version, python_full_version))
    for extra in data["project"]["optional-dependencies"]:
        for line in data["project"]["optional-dependencies"][extra]:
            base_constrains.update(parse_single_requirement(line, python_version, python_full_version))
    return base_constrains



# SPDX-FileCopyrightText: 2023-present Grzegorz Bokota <bokota+github@gmail.com>
#
# SPDX-License-Identifier: MIT
"""tox plugin for simplify minimal requirements tests by creating minimal constrains file."""
from tox_min_req._parse_dependencies import parse_pyproject_toml, parse_setup_cfg, parse_single_requirement
from tox_min_req._version import __version__

__all__ = ("__version__", "parse_setup_cfg", "parse_pyproject_toml", "parse_single_requirement")
